/*
 *   VPA10k -- VPA with 10000 Minefields
 *
 *   This program converts the "bigminefields" from the UTILx.DAT
 *   files into markers.
 *
 *   Yes, I know this code is ugly.
 *
 *   (c) 2001 Stefan Reuther <Streu@gmx.de>. No warranties.
 */
#include <cstring>
#include <string>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <vector>
#include <map>
#include <cmath>
#include <cctype>
#include "buffer.h"

/** Tag for our markers. There is an unused field in the
    marker struct, which we use to find the markers later.
    Change this to 0 if it makes problems, but then automatic
    deletion will no longer work. */
#define OUR_FUNKY_TAG 4711

/** Compile-time assert. Compiler fails with `array size negative'
    if assert violated. */
#define CTASSERT(cond) do { char a[(cond) ? 1 : -1]; } while(0)

#define VPA_SIG "VPA Database\015\012\006"
#define VPA_SIG_LEN (sizeof(VPA_SIG)-1)

using namespace std;

/** Minefield. */
struct Minefield {
    short id;
    short x, y, owner;
    long units;
    short type, fcode, event;
    short dummy;                 // padding which would otherwise be
                                 // inserted by compiler
};

/** VPA DB Chunk. */
class CVpaChunk {
 public:
    char name[4];            ///< Chunk name
    long size;               ///< Size (not necessarily up-to-date)
    short count;             ///< Object count
    CBuffer buf;             ///< Contents
};

/** Simple class for one VPA DB turn. */
class CVpaTurn {
 public:
    char header[20 + 44*2];                    ///< turn nr, score, etc.
    vector<CVpaChunk*> chunks;

    CVpaChunk* findChunk(const char* name);
    void write(ostream& out);
};

vector<CVpaTurn*> turns;     ///< The VPA database
map<int, Minefield> mines;   ///< The minefields
map<string, int> color;      ///< Maps color schemes to colors

char* dir;                   ///< Game directory
string dir_s;                ///< Game directory plus slash
char* player;                ///< Player (string)
char timestamp[18];          ///< Timestamp (not 0-terminated)
int player_colors[11];       ///< Maps player numbers to colors

#ifdef __DJGPP__
extern "C" {
char** __crt0_glob_function(char*) { return 0; }
void __crt0_load_environment_file(char*) { }
}
#endif

void
die (string s)
{
    cout << "ERROR: " << s << endl;
    exit(1);
}

/** Get named chunk. \param name 4-byte name \returns chunk or null */
CVpaChunk*
CVpaTurn::findChunk(const char* name)
{
    for (unsigned i = 0; i < chunks.size(); ++i)
        if (memcmp(name, chunks[i]->name, 4) == 0)
            return chunks[i];
    return 0;
}

/** Write turn to database file. */
void
CVpaTurn::write(ostream& out)
{
    /* compute size */
    long total = 0;
    for (unsigned i = 0; i < chunks.size(); ++i)
        total += 10 + (chunks[i]->size = chunks[i]->buf.Length());

    out.write("TURN", 4);
    out.write(&total, 4);
    out.write(header, sizeof(header));
    for (unsigned i = 0; i < chunks.size(); ++i) {
        out.write(chunks[i]->name, 4);
        out.write(&chunks[i]->size, 4);
        out.write(&chunks[i]->count, 2);
        out.write(chunks[i]->buf.Ptr(), chunks[i]->buf.Length());
    }
    if (!out)
        die ("write error");
}

/** Load one turn. \returns turn, or null if none left in file */
CVpaTurn*
loadVpaTurn(istream& in)
{
    long header[2];
    if (!in.read (&header, 8) || header[0] != 0x4E525554)
        return 0;

    CVpaTurn* t = new CVpaTurn;
    in.read (&t->header, sizeof(t->header));
    cout << "\tTurn " << string(t->header+2, 18) << "\n";
    while (header[1] > 0) {
        CVpaChunk* c = new CVpaChunk;
        in.read (&c->name, 4);
        in.read (&c->size, 4);
        in.read (&c->count, 2);
        header[1] -= 10 + c->size;
        if (header[1] < 0)
            die ("invalid file format");

        char* b = new char[c->size];
        in.read (b, c->size);
        c->buf.Append(b, c->size);
        delete[] b;

        t->chunks.push_back(c);
    }
    return t;
}

/** Load VPA Database. Constructs the #turns array. */
void
loadVpaDatabase()
{
    string name = dir_s + "vpa" + player + ".db";
    ifstream in (name.c_str(), ios::binary | ios::in);
    if (!in)
        die ("can't open " + name);
    cout << "Loading " << name << "...\n";

    char buffer[VPA_SIG_LEN];
    in.read(buffer, VPA_SIG_LEN);
    if (memcmp(buffer, VPA_SIG, VPA_SIG_LEN) != 0)
        die ("this is not a VPA database");

    /* Next, we read a TURN block */
    while (CVpaTurn* t = loadVpaTurn(in))
        turns.push_back(t);
    in.clear();
    cout << "\tTotal " << in.tellg() << " bytes read\n";
}

/** Read UTILx.DAT. Extracts relevant minefields, and the turn timestamp. */
void
loadUtilData()
{
    string name = dir_s + "util" + player + ".dat";
    ifstream in (name.c_str(), ios::binary | ios::in);
    if (!in)
        die ("can't open " + name);
    cout << "Loading " << name << "...\n";

    unsigned short head[2];
    int nmines = 0, nhigh = 0;
    while (in.read (&head, 4)) {
        long fpos = in.tellg();
        if (head[0] == 0) {
            Minefield mf;
            if (!in.read (&mf, sizeof(mf)-2))    // -2 due to padding
                die ("read error\n");
            mf.dummy = 0;
            nmines++;
            if (mf.id > 500) {
                mines[mf.id] = mf;
                nhigh++;
            }
        } else if (head[0] == 13) {
            in.read (timestamp, 18);
        }
        in.seekg(fpos + head[1]);
    }
    cout << "\t" << nmines << " minefield records, " << nhigh << " high id, "
            "thereof " << mines.size() << " unique\n";
}

/** Get turn for timestamp. \param ts 18-character timestamp.
    \returns turn, or null */
CVpaTurn*
findTurn (const char* ts)
{
    for (unsigned i = 0; i < turns.size(); ++i)
        if (memcmp(ts, turns[i]->header + 2, 18) == 0)
            return turns[i];
    return 0;
}

/** Color for specified player. Uses color from ini file, or default
    if none set. */
int
mineColor(int owner)
{
    if (owner <= 0 || owner > 11 || !player_colors[owner-1])
        return 4;  // red
    else
        return player_colors[owner-1];
}

/** Append all minefields to marker chunk c. Increases count
    accordingly. Assumes c does not yet contain the string table. */
void
addMinefields(CVpaChunk* c)
{
    for (map<int, Minefield>::iterator i = mines.begin(); i != mines.end(); ++i) {
        c->buf.AppendWord(7 + 256*mineColor(i->second.owner));
        c->buf.AppendWord(i->second.x);
        c->buf.AppendWord(i->second.y);
        c->buf.AppendWord(0);
        c->buf.AppendWord(OUR_FUNKY_TAG);
        c->buf.AppendWord(sqrt(i->second.units));
        c->count++;
    }
    cout << "\tTotal " << c->count << " markers\n";
}

/** Create new marker chunk. */
CVpaChunk*
makeMarkerChunk ()
{
    cout << "Creating new markers...\n";

    CVpaChunk* c = new CVpaChunk;
    memcpy(c->name, "MARK", 4);
    c->size = 0;                  // recomputed on save
    c->count = 0;
    addMinefields(c);
    return c;
}

/** Modify existing marker chunk. Removes all markers we have
    created, and adds minefield markers. */
void
frobMarkerChunk(CVpaChunk* c)
{
    CBuffer orig;
    int origc;

    cout << "Modifying markers...\n";

    orig.Append(c->buf.Ptr(), c->buf.Length());
    c->buf.Clear();
    origc = c->count;
    c->count = 0;

    /* now, the original marker chunk is empty. Copy markers. */
    int index = 0;
    while (origc-- > 0) {
        if (OUR_FUNKY_TAG == 0 || orig.GetWord(index + 8) != OUR_FUNKY_TAG) {
            /* copy it */
            c->buf.Append(&orig[index], 12);
            c->count++;
        }
        index += 12;
    }
    cout << "\t" << c->count << " markers carried over from previous turn\n";

    addMinefields(c);

    /* carry over text strings */
    c->buf.Append(&orig[index], orig.Length() - index);
}

/** Rewrite VPA database to disk. */
void
writeVpaDatabase()
{
    string name = dir_s + "vpa" + player + ".db";
    ofstream out (name.c_str(), ios::binary | ios::out);
    if (!out)
        die ("can't create " + name);
    cout << "Writing " << name << "...\n";

    out.write (VPA_SIG, VPA_SIG_LEN);
    for (unsigned i = 0; i < turns.size(); ++i)
        turns[i]->write(out);
}

/** Remove leading and trailing whitespace. */
void
chomp(string& s)
{
    while (!s.empty() && isspace(s[0]))
        s.erase(0, 1);
    while (!s.empty() && isspace(s[s.length()-1]))
        s.erase(s.length()-1);
}

/** Convert string to uppercase */
void upcase(string& s)
{
    for (unsigned i = 0; i < s.length(); ++i)
        s[i] = toupper(s[i]);
}

/** Load VPA.INI. This one defines color profiles. In each color profile,
    the last color is the minefield color. \returns true if file found,
    false if not. */
bool
loadVpaIni(string n)
{
    ifstream in(n.c_str());
    if (!in)
        return false;

    cout << "Loading " << n << "...\n";
    string s;
    while (getline(in, s)) {
        chomp(s);
        upcase(s);
        if (s.substr(0, 7) == "PROFILE") {
            string name;

            s.erase(0, 7);
            unsigned n = s.find_first_not_of(" \t=");
            if (n != string::npos)
                s.erase(0, n);
            n = s.find(',');
            name = s.substr(0, n);
            n = s.rfind(',');
            if (n != string::npos)
                color[name] = atoi(s.substr(n+1).c_str());
        }
    }
    return true;
}

/** Load VPACLRx.INI. This one assigns players to color schemes.
    \returns true if file found, false if not */
bool
loadVpaColors(std::string n)
{
    ifstream in (n.c_str());
    if (!in)
        return false;
    cout << "Loading " << n << "...\n";

    string s;
    while (getline(in, s)) {
        chomp(s);
        upcase(s);
        if (s.substr(0, 9) != "RACECOLOR")
            continue;

        int race = atoi(s.substr(9, 2).c_str());
        if (race <= 0 || race > 11)
            continue;

        unsigned x = s.find_last_of(" \t=");
        if (x != string::npos)
            player_colors[race-1] = color[s.substr(x+1)];
    }
    return true;
}

/** Load VPA INI files. To get colors right. */
void
loadIniFiles()
{
    if (!(loadVpaIni(dir_s + "vpa.ini") || loadVpaIni("vpa.ini"))) {
        cout << "warning: VPA.INI not found, assuming defaults\n";
        color["OWN"] = 3;
        color["ENEMY"] = 5;
        color["ALLY"] = 3;
    }
    loadVpaColors(dir_s + "vpaclr" + player + ".ini") ||
        loadVpaColors(string("vpaclr") + player + ".ini");
}

/** Friendly Help Message. */
void 
help ()
{
    cout << "VPA10K -- VPA with 10000 Minefields -- 2001 by Stefan Reuther <Streu@gmx.de>\n"
            "\n"
            "Usage: VPA10K <player> <dir>\n"
            "\n"
            "Preconditions: turn must be unpacked, VPA must have seen it\n"
            "i.e., before this program, do\n"
            "\tccunpack <params>\t\t;; let me plug just once ;)\n"
            "\tvpa <params> /b\n";
    exit (1);
}

int
main(int argc, char** argv)
{
    CTASSERT(sizeof(Minefield) == 20);
    while (*++argv) {
        if (strcmp (*argv, "--help") == 0 || strchr (*argv, '?') != 0) {
            help();
            return 0;
        } else if (!player) {
            player = *argv;
        } else if (!dir) {
            dir = *argv;
        } else
            die ("invalid command line");
    }
    if (!dir)
       die ("missing directory name");
    dir_s = dir;
    if (dir_s.length() && strchr("/\\:", dir_s[dir_s.length()-1]) == 0)
       dir_s += "/";

    loadVpaDatabase();
    loadUtilData();
    loadIniFiles();
    CVpaTurn* t = findTurn(timestamp);
    if (!t)
       die ("current turn not in VPA database. Remember unpack + vpa/b first");
    CVpaChunk* c = t->findChunk("MARK");
    if (!c)
        t->chunks.push_back(makeMarkerChunk());
    else
        frobMarkerChunk(c);
    writeVpaDatabase();
}
